-- ========================================================
--                          Data
-- ========================================================

PABR_ActiveProfile = "Default"

PABR_Profiles = { } -- Loaded by WoW

PABR_DefaultConfig = {
	Name = "Default",

	EnableTest = false,
	UILocked = false,
	
	ShowLocalPlayer = true,
	EnableInspect = true,
	EnableActivityFilter = true,
	ActivityTimeout = 60,
	ActivityDistance = 120,

	EnableArena = 3,
	EnableRatedBg = 3,
	EnableBg = 3,
	EnableRaid = 0,
	EnableParty = 0,
	EnableSolo = 0,

	ShowEffect = true,
	ShowPortrait = true,
	FlashTargetCooldowns = true,
	MaxCooldowns = 32,
	MaxCooldownsPerPlayer = 8,

	ShowAlert = 2,
	AlertSound = true,
	AlertThreshold = 180,
	AlertHoldTimer = 1.5,
	MaxAlerts = 5,
}

PABR_Config = { } -- Modified to point to the active profile

local VERIFIED = false

-- ========================================================
--                         Functions
-- ========================================================

local function verifySpells(config, reset)
	local spells = config.Spells
	if not spells then
		spells = { }
		config.Spells = spells
	elseif reset then
		wipe(spells)
	end

	-- Make sure new important spells are enabled
	local updated = false
	local list = PABR_GetTaggedSpells("important")
	for index, spellId in ipairs(list) do
		if spells[spellId] == nil then
			spells[spellId] = true
			updated = true
		end
	end
	return updated
end

local function verifyConfig(config, reset)
	if reset then
		for k, v in pairs(config) do
			if type(v) ~= "table" and k ~= "Name" then
				config[k] = nil
			end
		end
	end

	local updated = false
	for k, v in pairs(PABR_DefaultConfig) do
		if config[k] == nil then
			config[k] = v
			updated = true
		end
	end
	return updated
end

local function findProfile(name)
	for k, v in ipairs(PABR_Profiles) do
		if v.Name == name then return v, k; end
	end
	return nil, 0
end

local function checkActiveProfile()
	return PABR_SetActiveProfile(PABR_ActiveProfile) or PABR_SetActiveProfile("Default") or PABR_SetActiveProfile(PABR_Profiles[1].Name)
end

function PABR_IsConfigReady()
	return VERIFIED
end

function PABR_InitConfig()
	local result = "ok"
	if not VERIFIED then
		if #PABR_Profiles == 0 then
			local config = { }
			verifyConfig(config)
			verifySpells(config)
			tinsert(PABR_Profiles, config)
			result = "init"
		else
			for k, v in ipairs(PABR_Profiles) do
				if verifyConfig(v) then result = "update"; end
				if verifySpells(v) then result = "update"; end
			end
		end
		assert(checkActiveProfile())
		VERIFIED = true
	end
	return result
end

function PABR_SetActiveProfile(name)
	local profile = findProfile(name)
	if profile then
		PABR_ActiveProfile = profile.Name
		PABR_Config = profile
		return true
	end
	return false
end

function PABR_ResetActiveProfile(which) -- which can be either "config" or "spells"
	if which == "config" then
		verifyConfig(PABR_Config, true)
	elseif which == "spells" then
		verifySpells(PABR_Config, true)
	end
end

function PABR_CloneProfile(name, newName, execute)
	local source = findProfile(name)
	if source and #newName > 0 and not findProfile(newName) then
		if execute then
			local config = PABR_DeepCopy(source)
			config.Name = newName
			tinsert(PABR_Profiles, config)
		end
		return true
	end
	return false
end

function PABR_DeleteProfile(name, execute)
	local target, index = findProfile(name)
	if target and #PABR_Profiles > 1 then
		if execute then
			tremove(PABR_Profiles, index)
			assert(checkActiveProfile())
		end
		return true
	end
	return false
end

function PABR_GetProfileNames()
	local names = { }
	for i=1, #PABR_Profiles do
		tinsert(names, PABR_Profiles[i].Name)
	end
	return names
end

function PABR_WipeProfiles()
	VERIFIED = false
	wipe(PABR_Profiles)
	PABR_InitConfig()
end

function PABR_IsSpellShown(spellId)
	return VERIFIED and PABR_Config.Spells[spellId]
end

function PABR_SetSpellShown(spellId, shown)
	if VERIFIED then PABR_Config.Spells[spellId] = not not shown; end
end