local UNITS = { }
local REF = { }

local function registerUnit(unit, ref)
	tinsert(UNITS, unit)
	tinsert(REF, ref)
end

do
	registerUnit("player", "player")
	registerUnit("pet", "player")

	for i=1, 5 do
		registerUnit("arena"..i, "player")
		registerUnit("party"..i, "player")
	end
	for i=1, 40 do
		registerUnit("raid"..i, "player")
	end

	registerUnit("playertarget", "player")
	registerUnit("pettarget", "pet")
	registerUnit("target", "player")
	registerUnit("targettarget", "target")
	registerUnit("focus", "player")
	registerUnit("focustarget", "focus")
	registerUnit("mouseover", "player")
	registerUnit("mouseovertarget", "mouseover")

	for i=1, 5 do
		registerUnit("arena"..i.."target", "arena"..i)
		registerUnit("party"..i.."target", "party"..i)
	end
	for i=1, 40 do
		registerUnit("raid"..i.."target", "raid"..i)
	end
end

local DIRTY = false
local ENABLED = false

local offset = 0

local output = { {}, {} }

local callbacks = { { {}, {} }, { {}, {} } }

local function fireCallback(set, teamId, ...)
	local list = callbacks[set][teamId]
	for _, entry in ipairs(list) do
		entry[1](entry[2], teamId, ...)
	end
end

local function checkAlertFilter(guid)
	local filter = PABR_Config["ShowAlert"]
	if filter >= 3 then return true end
	if filter >= 2 and UnitGUID("focus") == guid then return true end
	if filter >= 1 and UnitGUID("target") == guid then return true end
	return false
end

-- API

function PABR_IsEngineEnabled()
	return ENABLED
end

function PABR_GetTeamNumCooldowns(teamId) -- return numCooldowns
	return #output[teamId]
end

function PABR_GetTeamCooldownInfo(teamId, index) -- return id, spellId, guid, name, cooldownStart, cooldownDuration, cooldownCount, cooldownMaxCount, effectStart, effectDuration
	return unpack(output[teamId][index])
end

function PABR_RegisterUpdateCallback(teamId, func, arg) -- arg, teamId arguments are passed to callback
	tinsert(callbacks[1][teamId], {func, arg})
end

function PABR_RegisterAlertCallback(teamId, func, arg) -- arg, teamId, guid, name, spellId, tags arguments are passed to callback
	tinsert(callbacks[2][teamId], {func, arg})
end

function PABR_FindUnit(guid) -- return unit, observerUnit
	if guid then
		for i=1, #UNITS do
			if UnitGUID(UNITS[i]) == guid then
				return UNITS[i], REF[i]
			end
		end
	end
	return nil, nil
end

function PABR_GetTime(timestamp)
	return timestamp + offset
end

-- Functions

local available = { }

function PABR_CreateTable()
	local table
	if #available > 0 then
		table = available[#available]
		tremove(available, #available)
	else
		table = { }
	end
	return table
end

function PABR_DeleteTable(table)
	wipe(table)
	tinsert(available, table)
end

function PABR_DeepCopy(t)
	local res = t
	if type(t) == "table" then
		res = {}
		for k, v in pairs(t) do
			if type(v) == "table" then
				v = PABR_DeepCopy(v)
			end
			res[k] = v
		end
	end
	return res
end

-- Handlers

function PABR_OnCombatLogEvent(timestamp, event, hideCaster, sourceGUID, sourceName, sourceFlags, sourceRaidFlags, destGUID, destName, destFlags, destRaidFlags, ...)
	if not PABR_IsEngineEnabled() or not sourceGUID or not destGUID then return; end

	if offset == 0 then
		offset = GetTime() - timestamp
	end

	if strsub(event, 1, 5) == "SPELL" then
		local spellId, spellName, spellSchool = ...

		if event == "SPELL_AURA_APPLIED" then
			PABR_OnEffectGain(timestamp, sourceGUID, destGUID, spellId)
			PABR_CheckCooldownDetection("effectup", timestamp, sourceGUID, destGUID, spellId)

		elseif event == "SPELL_AURA_REMOVED" then
			PABR_OnEffectFade(timestamp, sourceGUID, destGUID, spellId)
			PABR_CheckCooldownDetection("effectdown", timestamp, sourceGUID, destGUID, spellId)

  		elseif event == "SPELL_CAST_SUCCESS" then
			PABR_CheckCooldownDetection("cast", timestamp, sourceGUID, destGUID, spellId)

  		elseif event == "SPELL_SUMMON" then
			PABR_NotifySummon(sourceGUID, destGUID)
			PABR_CheckCooldownDetection("summon", timestamp, sourceGUID, destGUID, spellId)

		elseif event == "SPELL_DAMAGE" or event == "SPELL_HEAL" then
			PABR_CheckCooldownDetection("hit", timestamp, sourceGUID, destGUID, spellId)

		elseif event == "SPELL_MISSED" then 
			PABR_CheckCooldownDetection("miss", timestamp, sourceGUID, destGUID, spellId)

		elseif event == "SPELL_INTERRUPT" then
			local extraSpellId, extraSpellName = select(4, ...)
			PABR_HandleInterrupt(timestamp, sourceGUID, destGUID, spellId, extraSpellId)
		end

	elseif event == "UNIT_DIED" or event == "UNIT_DESTROYED" then
		PABR_HandleDeath(timestamp, destGUID)
	end
end

function PABR_ResetEngine()
	PABR_WipeEffects()
	PABR_WipeEnemyTeam()
	PABR_WipeSummon()
	PABR_WipeCooldowns()
	PABR_WipeInspect()
	PABR_WipeFilter()
end

function PABR_ToggleEngine(enable)
	ENABLED = enable
	PABR_UpdatePlayerTeam()

	-- Test alerts
	if enable and PABR_Config["EnableTest"] then
		for i=1, PABR_Config["MaxAlerts"] do
			fireCallback(2, 1, UnitGUID("player"), UnitName("player"), 42292, {"important", "defense"})
		end
	end
end

function PABR_SetDirty()
	DIRTY = true
end

function PABR_CheckRecompute()
	if DIRTY then
		DIRTY = false

		local localGUID = UnitGUID("player")
		local showLocal = PABR_Config["ShowLocalPlayer"]
		local hasFilter = PABR_IsFilterActive()
		local maxCooldownsPerPlayer = PABR_Config["MaxCooldownsPerPlayer"]

		for i=1, #output do
			local out = output[i]

			for k, v in pairs(out) do
				out[k] = nil

				PABR_DeleteTable(v)
			end

			if PABR_IsEngineEnabled() then
				if PABR_Config["EnableTest"] then
					for j=1, PABR_Config["MaxCooldowns"] do
						local t = PABR_CreateTable()
						t[1] = j
						t[2] = 42292
						t[3] = UnitGUID("player")
						t[4] = UnitName("player")
						t[5] = GetTime()
						t[6] = 60
						t[7] = 0
						t[8] = 0
						if math.fmod(j, 2) == 0 then
							t[9] = 0
							t[10] = 0
						else
							t[9] = GetTime()
							t[10] = 30
						end
						tinsert(out, t)
					end
				else
					for j=1, PABR_GetTeamSize(i) do
						local guid, name = PABR_GetTeamMember(i, j)
						local count = 0

						if (showLocal or guid ~= localGUID) and not (hasFilter and PABR_IsFiltered(guid)) then
							for k=1, PABR_GetNumCooldowns(guid) do
								local spellId, cooldownStart, cooldownDuration, count, maxCount, id = PABR_GetCooldownInfo(guid, k)

								if PABR_IsSpellShown(spellId) and count < maxCooldownsPerPlayer then
									local effectStart, effectDuration = PABR_GetEffectInfo(guid, spellId)

									local t = PABR_CreateTable()
									t[1] = id
									t[2] = spellId
									t[3] = guid
									t[4] = name
									t[5] = cooldownStart
									t[6] = cooldownDuration
									t[7] = count
									t[8] = maxCount
									t[9] = effectStart
									t[10] = effectDuration
									tinsert(out, t)
									
									count = count + 1
								end
							end
						end
					end
				end
			end

			fireCallback(1, i)
		end
	end
end

function PABR_CheckAlert(sourceGUID, spellId, cooldown, info)
	if checkAlertFilter(sourceGUID) and (PABR_Config["ShowLocalPlayer"] or sourceGUID ~= UnitGUID("player")) and not (PABR_IsFilterActive() and PABR_IsFiltered(sourceGUID)) then
		if PABR_IsSpellShown(spellId) and (cooldown >= PABR_Config["AlertThreshold"] or info.reset or info.trinket) then
			local teamId, _, _, sourceName = PABR_ResolveTeam(sourceGUID)
			if teamId then
				fireCallback(2, teamId, sourceGUID, sourceName, spellId, info.tags)
			end
		end
	end
end