local MAX_ICONS = 64
local MAX_ALERTS = 8

local ui = { }

local params = {{"Your team", {0, 1, 0}}, {"Enemy team", {1, 0, 0}}}

function PABR_IsMatching(id, enableMask)
	return (id == 0 and enableMask > 0) or bit.band(enableMask, 2 ^ (id - 1)) > 0
end

function PABR_IsEnabled(frame, enableMask)
	return PABR_IsMatching(frame:GetID(), enableMask)
end

function PABR_IsLocked(frame)
	return PABR_Config["UILocked"] or frame.manipulating
end

function PABR_ToggleUI(enableMask)
	for index, frame in pairs(ui) do
		if PABR_IsEnabled(frame, enableMask) then
			frame:Show()
		else
			frame:Hide()
		end
		if frame.Setup then
			frame:Setup(enableMask)
		end
	end
end

-- Services

local SOUND_SPAM = { }

function PABR_PlaySound(filename, spamTimer)
	local now = GetTime()
	if now > (SOUND_SPAM[filename] or 0) then
		PlaySoundFile("Interface\\AddOns\\PartyAbilityBarsReloaded\\sfx\\"..filename)
		SOUND_SPAM[filename] = now + (spamTimer or 0.5)
		--print("playing", filename)
	end
end

-- Container

function PABR_ContainerFrame_OnLoad(self)
	local id = self:GetID()
	local color = params[id][2]

	self.manipulating = false
	tinsert(ui, self)

	-- Drag config
	self:SetClampedToScreen(true)
	self:RegisterForDrag("LeftButton", "RightButton")

	-- Resize config
	self:SetMaxResize(580, 400)
	self:SetMinResize(80, 40)
	self:SetResizable(true)

	-- Set title
	self.title:SetText(params[id][1])
	self.title:SetTextColor(unpack(color))

	-- Register as callback
	PABR_RegisterUpdateCallback(id, PABR_ContainerFrame_Updated, self)

	-- Create items
	local items = { }
	for i=1, MAX_ICONS do
		tinsert(items, CreateFrame("Frame", self:GetName().."Item"..i, self, "PABR_ItemTemplate"))
	end
	self.items = items
	
	-- Setup logic
	self.Setup = function(self, enableMask)
		self.width = 0
		PABR_ContainerFrame_Rearrange(self)
	end
end

function PABR_ContainerFrame_OnShow(self)
	PABR_ContainerFrame_Updated(self)
	PABR_ContainerFrame_OnUpdate(self, 0)
end

function PABR_ContainerFrame_OnUpdate(self, elapsed)
	PABR_ContainerFrame_Rearrange(self, false)
end

function PABR_ContainerFrame_OnDragStart(self, button)
	if not PABR_IsLocked(self) then
		self.manipulating = true
		if button == "LeftButton" then
			self:StartMoving()
		elseif button == "RightButton" then
			self:StartSizing("LEFT")
		end
	end
end

function PABR_ContainerFrame_OnDragStop(self)
	if self.manipulating then
		self.manipulating = false
		self:StopMovingOrSizing()
		PABR_ContainerFrame_Rearrange(self, true)
		ValidateFramePosition(self, 25)
	end
end

function PABR_ContainerFrame_OnEnter(self)
	if not PABR_IsLocked(self) then
		GameTooltip:SetOwner(self, "ANCHOR_CURSOR")
		GameTooltip:AddDoubleLine("Mouse button", "Command")
		GameTooltip:AddDoubleLine("Hold left", "Move the frame", 1, 1, 1, 1, 1, 1)
		GameTooltip:AddDoubleLine("Hold right", "Resize the frame", 1, 1, 1, 1, 1, 1)
		GameTooltip:Show()
	end
end

function PABR_ContainerFrame_OnLeave(self)
	GameTooltip:Hide()
end

function PABR_ContainerFrame_Updated(self)
	if not self:IsShown() then return; end

	-- Query API
	local id = self:GetID()
	local items = self.items
	local numCooldowns = min(#items, PABR_Config["MaxCooldowns"], PABR_GetTeamNumCooldowns(id))
	for i=1, numCooldowns do
		local id, spellId, guid, name, cooldownStart, cooldownDuration, cooldownCount, cooldownMaxCount, effectStart, effectDuration = PABR_GetTeamCooldownInfo(id, i)
		PABR_ItemFrame_Show(items[i], id, spellId, guid, name, cooldownStart, cooldownDuration, cooldownCount, cooldownMaxCount, effectStart, effectDuration)
	end
	for i=numCooldowns+1, #items do
		items[i]:Hide()
	end
end

function PABR_ContainerFrame_Rearrange(self, minimize)
	local width = self:GetWidth()
	local usefulWidth = 0
	if width and (minimize or self.width ~= width) then
		self.width = width

		local items = self.items
		local x, y = 0, 0
		local usefulHeight = 0
		for i=1, min(#items, PABR_Config["MaxCooldowns"]) do
			local itemWidth, itemHeight = items[i]:GetSize()
			if x + itemWidth > width then
				x = 0
				y = y + itemHeight
			end
			items[i]:SetPoint("TOPLEFT", self, "TOPLEFT", x, -(y + 12))
			x = x + itemWidth
			usefulWidth = max(usefulWidth, x)
			usefulHeight = max(usefulHeight, y + itemHeight)
		end
		self:SetHeight(usefulHeight + 12)
	end
	if minimize then self:SetWidth(usefulWidth + 1) end
end

-- Item

local BLINK_CYCLE = 1

function PABR_ItemFrame_OnLoad(self)
	local color = params[self:GetParent():GetID()][2]
	self.duration:SetTexture(unpack(color))
	self.duration:SetAlpha(0.5)
end

function PABR_ItemFrame_OnShow(self)
	PABR_ItemFrame_OnUpdate(self, 0)
end

function PABR_ItemFrame_OnUpdate(self, elapsed)
	local text = self.text
	local cdText = self.cooldown.text
	local texture = self.duration
	local duration = self.effectDuration
	if duration > 0 then
		local timeLeft = max(self.effectStart + duration - GetTime(), 0)
		local pulse = 0.75 + 0.25 * cos(GetTime() * 360)
		if timeLeft < 60 then
			text:SetText(ceil(timeLeft))
		else
			text:SetText("!")
		end
		text:SetTextColor(pulse, pulse, pulse)
		text:Show()
		texture:SetPoint("TOPLEFT", self, "BOTTOMLEFT", 2, 2 + (self:GetHeight() - 4) * timeLeft / duration)
		texture:Show()
		cdText:Hide()
	else
		local timeLeft = max(self.cooldownStart + self.cooldownDuration - GetTime(), 0)
		if not OmniCC and timeLeft < 60 then -- Let OmniCC handle the text itself if user has installed it
			cdText:SetText(ceil(timeLeft))
			cdText:Show()
		else
			cdText:Hide()
		end
		text:Hide()
		texture:Hide()
	end
	if UnitName("target") == self.name and GetTime() % 0.5 < 0.1 and PABR_Config["FlashTargetCooldowns"] then
		self:SetAlpha(0)
	else
		self:SetAlpha(1)
	end
end

function PABR_ItemFrame_OnEnter(self)
	local spellName = GetSpellInfo(self.spell)
	if spellName then
		GameTooltip:SetOwner(self, "ANCHOR_RIGHT")
		GameTooltip:AddDoubleLine(self.name, spellName, 1, 1, 1, 1, 1, 0)
		GameTooltip:Show()
	end
end

function PABR_ItemFrame_OnLeave(self)
	GameTooltip:Hide()
end

function PABR_ItemFrame_Show(self, id, spellId, guid, name, cooldownStart, cooldownDuration, cooldownCount, cooldownMaxCount, effectStart, effectDuration)
	if not PABR_Config["ShowEffect"] then effectDuration = 0; end

	local icon = PABR_GetSpellIcon(spellId, self:GetParent():GetID() == 2)
	self.icon:SetTexture(icon)
	self.count:SetText((cooldownMaxCount > 0 and cooldownCount) or "")

	-- A bit ugly, but does the trick.
	local modified = false
	if self:GetID() ~= id then
		self:SetID(id)
		modified = true
	end
	if self.cooldownStart ~= cooldownStart then
		self.cooldownStart = cooldownStart
		modified = true
	end
	if self.cooldownDuration ~= cooldownDuration then
		self.cooldownDuration = cooldownDuration
		modified = true
	end
	if self.cooldownCount ~= cooldownCount then
		self.cooldownCount = cooldownCount
		modified = true
	end
	if self.effectStart ~= effectStart then
		self.effectStart = effectStart
		modified = true
	end
	if self.effectDuration ~= effectDuration then
		self.effectDuration = effectDuration
		modified = true
	end
	if modified then
		CooldownFrame_SetTimer(self.cooldown, cooldownStart, cooldownDuration, effectDuration == 0, cooldownCount, cooldownMaxCount)
	end

	-- TODO: bind a button to a CD so it is the button that is moved, not the cd between buttons

	self.name = name
	self.spell = spellId
	self:Show()
end

-- Alert manager

function PABR_AlertManagerFrame_OnLoad(self)
	self.manipulating = false
	tinsert(ui, self)

	-- Drag config
	self:SetClampedToScreen(true)
	self:RegisterForDrag("LeftButton")

	-- Set title
	self.title:SetText("Alerts area")

	-- Register as callback
	for i=1, 2 do
		PABR_RegisterAlertCallback(i, PABR_AlertManagerFrame_Push, self)
	end

	-- Create items
	local alerts = { }
	for i=1, MAX_ALERTS do
		local alert = CreateFrame("Frame", self:GetName().."Alert"..i, self, "PABR_AlertTemplate")
		alert.background:SetAlpha(0.5)
		alert:SetPoint("TOP", self, "TOP", 0, 8 - i * 24)
		tinsert(alerts, alert)
	end
	self.alerts = alerts

	-- Setup logic
	self.Setup = function(self, enableMask)
		self.enableMask = enableMask

		if PABR_IsLocked(self) then
			self.header:Hide()
			self.title:Hide()
			self:EnableMouse(false)
		else
			self.header:Show()
			self.title:Show()
			self:EnableMouse(true)
		end
	end
end

function PABR_AlertManagerFrame_OnDragStart(self, button)
	if not PABR_IsLocked(self) then
		self.manipulating = true
		self:StartMoving()
	end
end

function PABR_AlertManagerFrame_OnDragStop(self)
	if self.manipulating then
		self.manipulating = false
		self:StopMovingOrSizing()
		ValidateFramePosition(self, 25)
	end
end

function PABR_AlertManagerFrame_OnEnter(self)
	GameTooltip:SetOwner(self, "ANCHOR_CURSOR")
	GameTooltip:AddLine("Hold left to move the frame.", 1, 1, 1)
	GameTooltip:AddLine("This becomes invisible when you lock the UI.")
	GameTooltip:Show()
end

function PABR_AlertManagerFrame_OnLeave(self)
	GameTooltip:Hide()
end

function PABR_AlertManagerFrame_Push(self, teamId, guid, name, spellId, tags)
	if not self:IsShown() or not PABR_IsMatching(teamId, self.enableMask) then return; end

	local alerts = self.alerts
	for i=1, min(#alerts, PABR_Config["MaxAlerts"]) do
		local alert = alerts[i]
		if not alert:IsShown() then
			local color = params[teamId][2]
			local text = ("%s   < %s >"):format(name, GetSpellInfo(spellId))
			local sound = "alertGeneric.mp3"
			local role = nil
			if tContains(tags, "attack") then
				sound = "alertOffensive.mp3"
				role = "DAMAGER"
			elseif tContains(tags, "defense") then
				sound = "alertDefensive.mp3"
				role = "TANK"
			end
			
			alert.timer = 0
			alert.background:SetVertexColor(unpack(color))
			alert.icon:SetTexture(PABR_GetSpellIcon(spellId, teamId == 2))
			if role then
				alert.category:SetTexCoord(GetTexCoordsForRole(role))
				alert.category:Show()
			else
				alert.category:Hide()
			end
			alert.text:SetText(text)
			alert:Show()
			PABR_AlertFrame_OnUpdate(alert, 0)
			if PABR_Config["AlertSound"] then
				PABR_PlaySound(sound, 2)
			end
			
			break
		end
	end
end

-- Alert

function PABR_AlertFrame_OnUpdate(self, elapsed)
	local FADEIN_TIMER = 0.3
	local HOLD_TIMER = FADEIN_TIMER + PABR_Config["AlertHoldTimer"]
	local FADEOUT_TIMER = HOLD_TIMER + 0.3
	
	local timer = self.timer + elapsed
	if timer < FADEIN_TIMER then
		self:SetAlpha(timer / FADEIN_TIMER)
		self:SetScale(2 - 1 * timer / FADEIN_TIMER)
	elseif timer < HOLD_TIMER then
		self:SetAlpha(1)
		self:SetScale(1)
	elseif timer < FADEOUT_TIMER then
		self:SetAlpha(1 - (timer - HOLD_TIMER) / (FADEOUT_TIMER - HOLD_TIMER))
		self:SetScale(1)
	else
		self:Hide()
	end
	self.timer = timer
end